#!/bin/env python3

"""Tool to replace alternative spellings of native characters
(e.g. German umlauts [ä, ö, ü] etc. [ß]) with the proper native characters.
For example, this problem occurs when no proper keyboard layout was available.

This program is dictionary-based to check if replacements are valid words.

By default, reads from STDIN and writes to STDOUT.
"""

import argparse
import logging
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Union

try:
    import tkinter as tk

    _TKINTER_AVAILABLE = True
except ImportError:
    _TKINTER_AVAILABLE = False

try:
    import pyperclip

    _PYPERCLIP_AVAILABLE = True
except ImportError:
    _PYPERCLIP_AVAILABLE = False

from betterletter import _PACKAGE_ROOT, substituters
from betterletter.io import backup_clipboard, get_dictionary, get_language_mappings
from betterletter.iteration import splitlines

logger = logging.getLogger(__name__)


def parse(description: str, lang_choices: Iterable[str]) -> dict[str, Union[bool, str]]:
    """Prepares, runs and returns parsing of CLI arguments for the script."""
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "language",
        help="Text language to work with, in ISO 639-1 format.",
        choices=lang_choices,
    )

    kwargs: dict[str, Union[str, bool]]  # Satisfy mypy
    if _PYPERCLIP_AVAILABLE:
        kwargs = {
            "help": "Read from and write back to clipboard instead of STDIN/STDOUT.",
            "action": "store_true",
        }
    else:
        kwargs = {
            "help": "OPTION IGNORED (not available): package 'pyperclip' couldn't be imported.",
            # False in all cases:
            "action": "store_false",
            "default": False,
        }

    parser.add_argument("-c", "--clipboard", **kwargs)  # type: ignore

    parser.add_argument(
        "-f",
        "--force",
        help="Force substitutions and return the text version with the maximum number"
        " of substitutions, even if they are illegal words (useful for names).",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--reverse",
        help="Reverse mode, where all native characters are simply replaced"
        " by their alternative spellings.",
        action="store_true",
    )

    if _TKINTER_AVAILABLE:
        kwargs = {
            "help": "Stop and open a GUI prompt for confirmation before finishing.",
            "action": "store_true",
        }
    else:
        kwargs = {
            "help": "OPTION IGNORED (not available): package 'tkinter' couldn't be imported.",
            # False in all cases:
            "action": "store_false",
            "default": False,
        }

    parser.add_argument("-g", "--gui", **kwargs)  # type: ignore

    parser.add_argument(
        "-d",
        "--diff",
        help="Print a diff view of the substitutions to stderr.",
        action="store_true",
    )
    parser.add_argument(
        "--debug",
        help="Output detailed logging information.",
        action="store_true",
    )
    return vars(parser.parse_args())


def main() -> None:
    language_mappings = get_language_mappings()

    args = parse(description=__doc__, lang_choices=language_mappings.keys())

    if args["debug"]:
        # Leave at default if no logging/debugging requested.
        logging.basicConfig(level="DEBUG")

    language = str(args["language"])
    language_mapping = language_mappings[language]

    use_clipboard = args["clipboard"]

    if use_clipboard:
        in_text = pyperclip.paste()
        backup_clipboard(in_text, file=_PACKAGE_ROOT / Path(".clip.bak"))
        possible_empty_reason = "clipboard empty or binary (document, image, ...)"
    else:
        in_text = sys.stdin.read()
        possible_empty_reason = "STDIN was empty"

    if not in_text.strip():
        logger.warning(f"Empty input received ({possible_empty_reason}).")

    if args["reverse"]:
        out_text = substituters.backward(in_text, language_mapping)
    else:
        try:
            known_words = get_dictionary(
                language=language,
                letter_filters=language_mapping.keys(),
            )
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Dictionary for {language=} not available (looked for '{e.filename}')"
            ) from e

        out_text = "".join(
            substituters.forward(
                text=in_text,
                language_mapping=language_mapping,
                known_words=known_words,
                force=bool(args["force"]),
            )
        )

    if args["diff"] or args["gui"]:
        from difflib import Differ

        raw_diff = Differ().compare(splitlines(in_text), splitlines(out_text))
        # See https://docs.python.org/3/library/difflib.html#difflib.Differ :
        # a 'line common to both sequences' starts with two spaces -> we don't care
        diff = "".join(line for line in raw_diff if not re.match("^  ", line))

    if args["gui"]:
        root = tk.Tk()

        usage = tk.Label(
            text="ENTER to confirm, ESC to abort.",
            font="TkFixedFont",
            fg="white",
            bg="SlateGray",
        )
        usage.pack()

        difflabel = tk.Label(text=diff, font="TkFixedFont", justify="left")
        difflabel.pack()

        root.title("Diff view")

        # https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime
        if TYPE_CHECKING:
            Event = tk.Event[Any]
        else:
            Event = tk.Event

        def destroy(event: Event) -> None:
            """Destroys a window and ends mainloop."""
            root.destroy()

        def terminate(event: Event) -> None:
            destroy(event)
            sys.exit(1)

        root.bind("<Return>", destroy)  # Enter key
        root.bind("<Escape>", terminate)

        root.mainloop()

    if args["diff"]:
        sys.stderr.write(diff)

    if use_clipboard:
        pyperclip.copy(out_text)
    else:
        sys.stdout.write(out_text)


if __name__ == "__main__":
    main()



import json
import logging
from functools import partial
from pathlib import Path
from typing import Iterable, Optional

from betterletter import _RESOURCES, ISO6391Code, LanguageMapping, WordLookup
from betterletter.iteration import apply_to_all, filter_strings

logger = logging.getLogger(__name__)


# Be explicit about encoding, since Windows might default to something other than UTF8.
# Provide this globally so no spot is forgotten.
_ENCODING = "utf8"
open_with_encoding = partial(open, encoding=_ENCODING)


def read(file: Path) -> list[str]:
    with open_with_encoding(file) as f:
        lines = f.read().splitlines()
    logger.debug(f"Fetched list containing {len(lines)} items from {file}")
    return lines


def write(file: Path, lines: list[str]) -> None:
    with open_with_encoding(file, "w") as f:
        f.write("\n".join(lines))
    logger.debug(f"Wrote file containing {len(lines)} lines to {file}")


def get_dictionary(
    language: ISO6391Code,
    letter_filters: Optional[Iterable[str]] = None,
) -> WordLookup:
    """Provides words from a pre-processed file or additionally creates it if not found.

    Args:
        file: The pre-processed, line-separated file of items.
        fallback_file: File to use and create new file from if main file not found.
        letter_filters: List of substrings items in the fallback file must contain to
            be returned and written to a new, processed file.

    Returns:
        Collection of words from the read pre-preprocessed file.
    """
    base_path = _RESOURCES / Path("dicts")
    base_file = Path(language).with_suffix(".txt")
    file = base_path / Path("filtered") / base_file

    if letter_filters is None:
        letter_filters = []
    try:
        items = read(file)
        logger.debug("Found pre-processed list.")
    except FileNotFoundError:
        logger.warning("No pre-processed dictionary found, creating from original.")
        logger.warning("This only needs to be done once after package installation.")
        fallback = base_path / base_file
        items = read(fallback)
        logger.debug(f"Fetched unprocessed list.")
        items = list(filter_strings(items, letter_filters))
        write(file, items)
    return set(items)


def backup_clipboard(text: str, file: Path) -> None:
    """Writes text content to file for backup purposes."""
    with open_with_encoding(file, "w") as f:
        f.write(text)


def get_language_mappings() -> dict[ISO6391Code, LanguageMapping]:
    with open_with_encoding((_RESOURCES / "languages").with_suffix(".json")) as f:
        # Enforce lowercase so we can rely on it later on. Do not use `casefold` on keys,
        # it lowercases too aggressively. E.g., it will turn 'ß' into 'ss', while keys
        # are supposed to be the native letters themselves.
        language_mappings: dict[ISO6391Code, LanguageMapping] = apply_to_all(
            json.load(f), str.lower
        )
    return language_mappings



import logging
import re
from typing import Iterator

from betterletter import LanguageMapping, NativeSpelling, Span, WordLookup
from betterletter.iteration import all_lengths_combinations, subset

logger = logging.getLogger(__name__)


def _substitute_spans(
    string: str,
    substitutions: dict[Span, NativeSpelling],
) -> str:
    """Substitutes elements at given positions in a string.

    We cannot simply use `str.translate`, because the translation of e.g. 'ue' to 'ü'
    would be done on the entire string at once. There exists words for which this is not
    suitable, e.g. 'Kuechenfeuer' -> 'Küchenfeuer': two 'ue', only one of which is to be
    replaced.

    Args:
        string: The original string to substitute elements in.
        substitutions: A mapping of spans to which element (e.g. a letter) to insert at
            that position, replacing the original.

    Returns:
        The original string with all characters at the given spans replaced according to
        the passed mapping.
    """
    # Reverse to work back-to-front, since during processing, indices change as the
    # string changes length. If starting from the back, this issue is avoided; if
    # starting from the front (non-reverse sorting), later positions would get shifted.
    reverse_spans = sorted(substitutions.keys(), reverse=True)

    chars = list(string)  # Make mutable
    for span in reverse_spans:
        chars[slice(*span)] = substitutions[span]
    new_string = "".join(chars)
    logger.debug(f"Turned '{string}' into '{new_string}'.")
    return new_string


def forward(
    text: str,
    language_mapping: LanguageMapping,
    known_words: WordLookup,
    force: bool,
) -> Iterator[str]:
    """Substitutes all alternative spellings with their native versions in a text.

    Certain words can be problematic, e.g. 'Abenteuer'. The algorithm finds 'ue', but a
    possible substitution of 'Abenteür' is illegal. Therefore, words with replacements
    are only *candidates* at first and have to get checked against a dictionary.

    Further, for example, 'Abenteuerbuecher' contains two 'ue' elements, spanning (6, 8)
    and (10, 12), respectively (as returned by `re.Match.span`). We cannot (easily)
    determine algorithmically which spans would be the correct ones to substitute at.
    Therefore, get all possible combinations; in this example: the first, the second,
    and both (i.e., all combinations of all possible lengths).

    Args:
        text: The text in which alternative spellings are to be replaced.
        language_mapping: A mapping of native characters to their alternative
            spellings.
        known_words: A collection of known words against which potential candidates for
            replacement are compared (since by far not all replacements possible as
            per the mapping are also legal words).
        force: Whether to force substitutions. If yes, will return the version of the
            text with the most substitutions, regardless if these are legal words as per
            the list of known words. This is useful for names.
            It is assumed that the *shortest* possible word is the one with the most
            substitutions, since substitutions generally shorten words (like ue -> ü).

    Returns:
        The text with all alternative spellings of native characters in words replaced
        with (legal) versions with proper (UTF-8) characters.
    """
    native_to_alternative_regexes = {
        native: re.compile(alternative, re.IGNORECASE)
        for native, alternative in language_mapping.items()
    }

    items = re.split(r"(\w+)", text)  # Can be words and non-words ("!", "-", ...)

    for item in items:
        logger.debug(f"Processing {item=}")

        # Using spans as keys is valid, as they are unique and immutable.
        substitutions: dict[Span, NativeSpelling] = {}

        for native, alt_regex in native_to_alternative_regexes.items():
            for match in re.finditer(alt_regex, item):
                logger.debug(f"Found a {match=} in {item=}.")
                if any(letter.isupper() for letter in match.group()):
                    # Treat e.g. 'Ae', 'AE', 'aE' as uppercase 'Ä'
                    native = native.upper()
                    logger.debug(f"Treating {match=} as uppercase.")
                else:
                    # Reset to lowercase: might still be uppercase from last iteration.
                    native = native.lower()
                    logger.debug(f"Treating {match=} as lowercase.")
                substitutions[match.span()] = native

        if not substitutions:
            logger.debug(f"No substitutions to be made for {item=}, skipping.")
            yield item
            continue

        all_spans = list(all_lengths_combinations(substitutions))
        all_substitutions = [subset(substitutions, span) for span in all_spans]
        logger.debug(f"Will test the following substitution sets: {all_substitutions}")

        candidates = [
            _substitute_spans(item, substitution) for substitution in all_substitutions
        ]
        logger.debug(f"Word candidates for replacement are: {candidates}")

        if force:
            # There exists only one word with "the most substitutions"; all others have
            # fewer. There is no ambiguity as long as the mapping of alternative
            # spellings to originals is bijective, e.g. 'ue' only maps to 'ü' and vice
            # versa. This is assumed to always be the case.
            item = sorted(candidates, key=len)[0]
        else:
            # Check whether different, specific cases are legal words.
            if item.islower():
                # Lowercase stays lowercase: 'uebel' -> 'übel', but legally capitalized
                # words are assumed to never occur lowercased ('aepfel' is never
                # replaced, to e.g. 'Äpfel')
                case_checks = [str.lower]
            else:
                # Includes all-caps, capitalized, title or mixed case. All-caps words
                # ('AEPFEL') is to be replaced correctly ('ÄPFEL'); however, 'ÄPFEL'
                # won't be part of a lookup dictionary. So check the capitalized version
                # ('Äpfel' -> legal) instead. Similarly for legally lowercased words,
                # e.g. 'UEBEL': 'Übel' -> illegal, but 'übel' -> legal, so also check
                # `lower`.
                case_checks = [str.lower, str.capitalize]

            candidate_cases = {
                candidate: [case_check(candidate) for case_check in case_checks]
                for candidate in candidates
            }

            try:
                item = next(  # Get the first one, then quit
                    candidate
                    for candidate, cases in candidate_cases.items()
                    if any(case in known_words for case in cases)
                )
            except StopIteration:  # Not a problem, `item` remains untouched
                logger.info(
                    f"No dictionary entry found for any candidates in {candidate_cases}"
                )
        logger.debug(f"Yielding fully processed {item=}")
        yield item


def backward(
    text: str,
    language_mapping: LanguageMapping,
) -> str:
    """Replaces all native characters in a text with their alternative spellings.

    No dictionary check is performed, since this operation is much simpler than its
    reverse.

    Args:
        text: The string in which to do the substitutions.
        language_mapping: A mapping of native characters to their alternative spellings
            (like 'ä' -> 'ae'). Can be lower- or uppercase.

    Returns:
        The text with replacements made, in the correct case.
    """
    table = {}
    for native, alt in language_mapping.items():
        lower, upper = native.lower(), native.upper()
        table[lower] = alt.lower()
        table[upper] = alt.capitalize()

    # In the case of German, `"ß".upper()` will yield "SS" instead of "ẞ". However,
    # for `str.translate`, the table keys need to be of length one. Hence, correct
    # Python's Unicode replacement before proceeding.
    corrections = {
        "SS": "ẞ",
    }
    for original, correction in corrections.items():
        try:
            table[correction] = table.pop(original)
        except KeyError:  # If this changes in the future, allow failure here.
            pass

    # `str.translate` requires 'ordinals: string' mappings, not just 'string: string'
    trans = str.maketrans(table)
    return text.translate(trans)
